<#

.SYNOPSIS
	Scan Cerberus Native users & groups and check for RSA public keys with a weak exponent

.DESCRIPTION
	For each Cerberus Native user/group, check if a ssh public key exists in a format that could be generated by an old version of PuttyGen with a weak exponent

.EXAMPLE
	PS> scan_native_pubkeys.ps1

#>

$cfsdata = [Environment]::GetFolderPath('CommonApplicationData') + '\Cerberus LLC\Cerberus FTP Server\'

$minexp = 65537                               # minimum "safe" RSA exponent
$littleend = [bitconverter]::IsLittleEndian

$ssh2_pem_pattern = '---- BEGIN SSH2 PUBLIC KEY ----\s(?:\s*Comment:\s*"[^"]*"\s*)?(.*?)\s---- END SSH2 PUBLIC KEY ----'
$openssh_pattern  = 'ssh-rsa ([^ ]+)(?:\s,+)?\s?'

$sshkey = '00 00 00 07 73 73 68 2d 72 73 61'  # 7 chars = ssh-rsa

Function GetWithEndianness ([Byte[]] $bytes, [Int] $offset, [Int] $len)
{
  [byte[]] $array = $bytes[$offset..($offset+$len-1)]

  if ($littleend) {
    [Array]::Reverse($array)    # in-place
  }

  Write-Output -noEnumerate $array    # return Byte[] not Object
}

Function Check-SSH ([String] $type, [String] $entity, [String] $filename, [Int] $entry, [Byte[]] $bytes)
{
  $hexString = ($cert[0..10] | ForEach-Object ToString X2) -join ' '
  if ($hexString -eq $sshkey) {

    $explenbytes = GetWithEndianness $cert 11 4
    $explen = [bitconverter]::ToInt32($explenbytes,0)
  
    $expbytes = GetWithEndianness $cert 15 $explen
    $expbytes += (0,0,0,0)    # make sure we're at least 4 bytes long
  
    $exponent = [bitconverter]::ToInt32($expbytes,0)
  
    if ($exponent -lt $minexp) {
      Write-Host "`t " $type " " -NoNewLine
      Write-Host $entity -NoNewLine -ForegroundColor Yellow
      Write-Host " Cert Entry" $count -NoNewLine
      Write-Host ' fails check' -NoNewLine -ForegroundColor Red
      Write-Host ' with a RSA exponent of' $exponent
    }
  }
}

Function Check-SSH2-PEM-PubKey ([String] $type, [String] $entity, [String] $filename, [String] $filedata)
{
  $count = 0
  $result = [regex]::Matches($filedata, $ssh2_pem_pattern, 'Singleline,IgnoreCase')
  
  Foreach ($base64 in $result)
  {
    ++$count
    
    $cert = [System.Convert]::FromBase64String($base64.Groups[1])
    Check-SSH $type $entity $filename $count $cert
  }
}

Function Check-OpenSSH-PubKey ([String] $type, [String] $entity, [String] $filename, [String[]] $filelines)
{
  $count = 0

  ForEach ($filedata in $filelines)
  {
    ++$count
    $result = [regex]::Matches($filedata, $openssh_pattern, 'IgnoreCase')
  
    Foreach ($base64 in $result)
    {
      $cert = [System.Convert]::FromBase64String($base64.Groups[1])
      Check-SSH $type $entity $filename $count $cert
    }    
  }
}

Function Get-UserFileList
{
  $maxver = 0
  $filename = ''
  Get-ChildItem -Path $cfsdata |
  	Where-Object { $_.Name -match '^users_(\d+\.\d+)\.xml$' } |
  	Foreach { $ver = [float]$Matches[1]; if ($ver -gt $maxver){ $maxver = $ver; $filename = $Matches[0] }}

  $userfile = $cfsdata + $filename

  $filelist = @()

  if ( (Test-Path -Path $userfile -PathType Leaf) ) {

    Write-Host "`nOpening Cerberus User file:" $userfile

    [XML]$xml = Get-content -Path $userfile

    $nshash = @{ ns1='http://cerberusllc.com/common' }
    # Find user who has a publicKey
    $xml | Select-Xml -Xpath '/cerberus/ns1:profiles/ns1:user[.//ns1:publicKey]' -Namespace $nshash | Select-Object -ExpandProperty Node | ForEach-Object {

      $username = $_.name 

      # For this user, get the publicKey
      Select-Xml -Xml $_ -Xpath './/ns1:publicKey[@contentType="filePath"]' -Namespace $nshash | Select-Object -ExpandProperty Node | ForEach-Object {
        $filelist += , @( "User", $username, $_.'#text' )   # note: comma forces array of arrays
      }
    }
  } else {
    Write-Host "Users" $userfile "Not Found" -ForegroundColor Red
  }
  
  $filelist
}

Function Get-GroupFileList
{
  $maxver = 0
  $filename = ''
  Get-ChildItem -Path $cfsdata |
  	Where-Object { $_.Name -match '^groups_(\d+\.\d+)\.xml$' } |
  	Foreach { $ver = [float]$Matches[1]; if ($ver -gt $maxver){ $maxver = $ver; $filename = $Matches[0] }}

  $groupfile = $cfsdata + $filename

  $filelist = @()

  if ( (Test-Path -Path $groupfile -PathType Leaf) ) {

    Write-Host "`nOpening Cerberus Group file:" $groupfile

    [XML]$xml = Get-content -Path $groupfile

    $nshash = @{ ns1='http://cerberusllc.com/common' }
    # Find user who has a publicKey
    $xml | Select-Xml -Xpath '/cerberus/ns1:groups/ns1:group[.//ns1:publicKey]' -Namespace $nshash | Select-Object -ExpandProperty Node | ForEach-Object {

      $groupname = $_.name 

      # For this group, get the publicKey
      Select-Xml -Xml $_ -Xpath './/ns1:publicKey[@contentType="filePath"]' -Namespace $nshash | Select-Object -ExpandProperty Node | ForEach-Object {
        $filelist += , @( "Group", $groupname, $_.'#text' )   # note: comma forces array of arrays
      }
    }
  } else {
    Write-Host "Groups" $groupfile "Not Found" -ForegroundColor Red
  }
  
  $filelist
}

Function Check-FIPS
{
  $fipsenabled = $false
  
  $settingsfile = $cfsdata + 'settings.xml'
  if ( (Test-Path -Path $settingsfile -PathType Leaf) ) {

    Select-Xml -LiteralPath $settingsfile -Xpath '/cerberus/settings/secure/enableFIPS140-2' | Select-Object -ExpandProperty Node | ForEach-Object {
      $fipsenabled = ($_.'#text' -eq 'true')
    }
  } else {
    Write-Host "Settings" $settingsfile "Not Found" -ForegroundColor Red
  }
  
  $fipsenabled
}


if (Check-FIPS) {

  Write-Host "`nFIPS is enabled so affected users will be prevented from logging-in in the future" -ForegroundColor Yellow
} else {

  Write-Host "`nYou do NOT have FIPS enabled so no users will be prevented from logging-in" -ForegroundColor Green
}

$files = @(Get-UserFileList) + @(Get-GroupFileList)
Write-Host "`nFound" $files.length 'public keys to check'

ForEach ($data in $files)
{
  ($type, $entity, $file) = $data
  Write-Host "`nChecking:" $file

  if ( (Test-Path -Path $file -PathType Leaf) ) {

    $lines = Get-Content $file      # We could use -Raw to get entire file as single string, but then we'd have to read it twice
    $all_lines = $lines -join "`n"  # Join with a newline so regex works
    
    Check-SSH2-PEM-PubKey $type $entity $file $all_lines
    Check-OpenSSH-PubKey $type $entity $file $lines
  } else {
    Write-Host "`tFile Not Found for " $type " " $entity -ForegroundColor Red
  }
}

Write-Host
Exit $LASTEXITCODE
